@@ -783,11 +783,6 @@ w32_strerror (int ec)
 
   if (ec == -1)
     ec = (int)GetLastError ();
-#ifdef HAVE_W32CE_SYSTEM
-  /* There is only a wchar_t FormatMessage.  It does not make much
-     sense to play the conversion game; we print only the code.  */
-  snprintf (strerr, sizeof strerr, "ec=%d", (int)GetLastError ());
-#else
   FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM, NULL, ec,
                  MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
                  strerr, DIM (strerr)-1, NULL);
@@ -797,7 +792,6 @@ w32_strerror (int ec)
     if (n > 2 && strerr[n-2] == '\r' && strerr[n-1] == '\n' )
       strerr[n-2] = 0;
   }
-#endif
   return strerr;
 }
 #endif /*HAVE_W32_SYSTEM*/
@@ -1695,10 +1689,16 @@ format_text (const char *text_in, int target_cols, int max_cols)
 }
 
 
-/* Substitute environment variables in STRING and return a new string.
- * On error the function returns NULL.  */
+/* Substitute variables in STRING and return a new string.  GETVAL is
+ * a function which maps NAME to its value; that value is a string
+ * which may not change during the execution time of this function.
+ * If GETVAL returns NULL substitute_vars returns NULL and the caller
+ * may inspect ERRNO for the reason.  In all other error cases this
+ * function also returns NULL.  Caller must free the returned string.  */
 char *
-substitute_envvars (const char *string)
+substitute_vars (const char *string,
+                 const char *(*getval)(void *cookie, const char *name),
+                 void *cookie)
 {
   char *line, *p, *pend;
   const char *value;
@@ -1749,19 +1749,22 @@ substitute_envvars (const char *string)
         {
           int save = *pend;
           *pend = 0;
-          value = getenv (p+2);
+          value = getval (cookie, p+2);
           *pend++ = save;
         }
       else
         {
           int save = *pend;
           *pend = 0;
-          value = getenv (p+1);
+          value = getval (cookie, p+1);
           *pend = save;
         }
 
       if (!value)
-        value = "";
+        {
+          xfree (result);
+          return NULL;
+        }
       valuelen = strlen (value);
       if (valuelen <= pend - p)
         {
@@ -1797,3 +1800,26 @@ substitute_envvars (const char *string)
  leave:
   return result;
 }
+
+
+/* Helper for substitute_envvars.  */
+static const char *
+subst_getenv (void *cookie, const char *name)
+{
+  const char *s;
+
+  (void)cookie;
+
+  s = getenv (name);
+  return s? s : "";
+}
+
+
+/* Substitute environment variables in STRING and return a new string.
+ * On error the function returns NULL.  */
+char *
+substitute_envvars (const char *string)
+{
+  return substitute_vars (string, subst_getenv, NULL);
+
+}
